#ifndef _LIST_H_
#define _LIST_H_

#include <string>

/* A Linked List class: Nodes are dynamically allocated upon introduction of
 * new elements (via append and insert).  Elements in the list are treated as
 * homogenous, though inheritance can be used to get around this restriction.
 */
template <typename Elem>
class List {
private:
    /* A linked list is a series of Nodes.  In this case, the Node class serves
     * primarily as a wrapper class around a data element.  Each Node maintains
     * a connection to the Nodes neighboring it (denoted by prev and next).
     */
    class Node
    {
    private:
        Elem *mElem;
        Node *mNext;
        Node *mPrev;

        // mParent refers to the list that created this particular instance.
        List<Elem> *mParent;

        Node();
    public:
        explicit Node(List<Elem> *parent, Elem *elem);
        ~Node();

        void setNext(Node *next);
        Node * getNext();

        void setPrev(Node *prev);
        Node * getPrev();

        List<Elem> * getParent();
        Elem * getElem();

        // Herebe a copy constructor: syntax sugar for copying one instance of
        // a Node onto another.
        Node& operator = (const Node& that);
    };

    // Beginning of the list; contains the first element 
    Node *mHead; 
    // Marks the tail end of the list.  Note that this doesn't contain the last
    // element in the list, but a NULL element instead to mark the list's end.
    Node *mTail;        

    // Counter to keep track of how many elements are in the list.
    int mSize;
    
public:
    // The iterator class allows users to scroll through list elements in
    // linear fashion.
    class iterator
    {
    private:
        // List that spawned this iterator.  If only the default constructor
        // has been used, then mParent is set to NULL.
        List *mParent;

        // Cursor position in the parent list.
        Node *mNode;

        // When we declare a class as a friend: we give it public access to all
        // data members, public or private.
        friend class List;
        iterator(Node *node);

    public:
        iterator();
        ~iterator();

        // An invalid iterator is one that doesn't belong to any particular
        // list, i.e. mNode == NULL
        const bool invalid() const;

        // Fetches the list end.
        const iterator end() const;

    private:
        Node * getNode();
        const Node * getNode() const;
   
        void setNode(Node *node);

        List<Elem> * getParent();
        const List<Elem> * getParent() const;

    public:
        Elem * getElement();
        const Elem * getElement() const;

        // Unary operators for traversing the list:
        iterator& operator --();    // prefix
        iterator operator --(int);  // postfix

        iterator& operator ++();    // prefix
        iterator operator ++(int);  // postfix

        iterator& operator = (const iterator& that);

        // Comparison operators
        const bool operator != (const iterator& that) const;
        const bool operator == (const iterator& that) const;
    };

    // Constructor:
    // The keyword 'explicit' just means we can't create a List via
    //      Elem *elem = new Elem();
    //      List myList = elem;
    // If no argument to the constructor is specified, then by default we set
    // node to NULL.
    explicit List(Elem *elem=NULL);
    // Destructor:
    ~List();

    // Inserts an element at the current position in our list
    void insert(iterator& it, Elem *node);

    // Appends an element to the end of our list:
    void append(Elem *node);

    // Removes an element at the iterator's position in our list:
    Elem * remove(iterator& it);

    // Return the size of our list:
    const int getSize() const;

    // Clears list contents.
    void clear();

private:
    friend class iterator;
    Node * getTail();
    const Node * getTail() const;

    Node * getHead();
    const Node * getHead() const;

public:
    iterator getIterator();
};

// Since this class definition is templated: the source definition needs to be
// included in the header file as well.
#include "List.hpp"
#include "ListNode.hpp"
#include "ListIterator.hpp"

#endif
