
// Constructor - if the default constructor is called, then elem is set to NULL
template <typename Elem>
List<Elem>::List(Elem *elem)
{
    // A list always has a tail, regardless of how many elements there are
    mTail = new Node(this, NULL);
    if (elem != NULL) {
        // Allocate a new node for the list head, and link to the tail
        mHead = new Node(this, elem);
        mTail->setPrev(mHead);

        // Update the size of the list
        mSize=1;
    } else {
        // With no element provided: we set the head to the tail
        mHead = mTail;

        // Don't forget to initialize the list size to zero!
        mSize = 0;
    }
    
}

// Destructor - called when the list is deleted.  Note that the actual elements
// are not deleted, only nodes.
template <typename Elem>
List<Elem>::~List()
{
    // Go through each node and delete
    Node *iter = mHead;
    while (iter != mTail)
    {
        Node *next = iter->getNext();
        delete iter;
        iter = next;
    }

    // A tail is allocated regardless of how many elements there are:
    // so be sure to delete it:
    delete mTail;
}

// Insert method: Inserts an element immediately before the cursor [e.g.
// iterator] position.
template <typename Elem>
void List<Elem>::insert(typename List<Elem>::iterator& it, Elem *elem)
{
    if (it.invalid()) return;

    // Check if the iterator belongs to the list
    Node *nit = it.getNode();
    if (nit->getParent() != this) return;

    // If the list size is zero, then allocate a new head.
    if (mSize == 0) {
        mHead = new Node(this, elem);
        mHead->setNext(mTail);
        mTail->setPrev(mHead);
        // Update iterator position
        it = typename List<Elem>::iterator(mHead);
    } else {
       
        // If the cursor is at the head, then we need to define a new head.
        if (nit == mHead) {
            mHead = new Node(this, elem);
            mHead->setNext(nit);
            nit->setPrev(mHead);
            it = typename List<Elem>::iterator(mHead);
        } else {
            // Get the previous node
            Node *prev = nit->getPrev();
            // Allocate a new node
            Node *node = new Node(this, elem);

            // Update link connections:
            node->setPrev(prev);
            prev->setNext(node);
            node->setNext(nit);
            nit->setPrev(node);

            // Update cursor position
            it.setNode(nit);
        }
    }

    // Size update!
    mSize++;
}

// Appends an element to the end of the list:
template <typename Elem>
void List<Elem>::append(Elem *elem)
{
    // Initialize Head given an empty list:
    if (mSize == 0) {
        mHead = new Node(this, elem);
        mHead->setNext(mTail);
        mTail->setPrev(mHead);

    } else {
        typename List<Elem>::iterator it(mTail);
        insert(it, elem);
    }

	// Size update!
	mSize++;
}

// Removes element from cursor position
template <typename Elem>
Elem * List<Elem>::remove(iterator& it)
{
    // Check if iterator is valid
    if (it.getParent() != this) return NULL;
    Node *node = it.getNode();
    if (node == mTail) return NULL;

    // An empty list has nothing to remove:
    if (mSize == 0) {
        return NULL;
    } else {
        // Delete the current node and update the link connections
        Elem *elem = node->getElem();
        Node *prev = node->getPrev();
        Node *next = node->getNext();

        delete node;

        // We define special cases depending on whether the cursor is pointed
        // at the head, tail, or if there's only one element in the list...
        if (mSize == 1) {
            mHead = mTail;
            mTail->setPrev(NULL);
            it.setNode(mTail);
        } else if (node == mHead) {
            next->setPrev(NULL);
            it.setNode(next);
            // mHead = next;
        } else if (next == mTail) {
            prev->setNext(mTail);
            mTail->setPrev(prev);
            it.setNode(prev);
        } else {
            prev->setNext(next);
            next->setPrev(prev);
            it.setNode(next);
        }

        // Size update!
        mSize--;
        return elem;
    }
}

template <typename Elem>
const int List<Elem>::getSize() const
{
    return mSize;
}

// Delete all elements from the list:
template <typename Elem>
void List<Elem>::clear()
{
    Node *cursor = mHead;
    while (cursor != NULL) {
        Node *next = cursor->getNext();
        delete cursor;
        cursor = next;
    }

    mHead = NULL;
    mTail->setPrev(mHead);
    mSize = 0;
}

template <typename TYPE>
typename List<TYPE>::Node*
List<TYPE>::getTail()
{
    return mTail;
}

template <typename TYPE>
const typename List<TYPE>::Node*
List<TYPE>::getTail() const
{
    return mTail;
}

template <typename TYPE>
typename List<TYPE>::Node*
List<TYPE>::getHead()
{
    return mHead;
}

template <typename TYPE>
const typename List<TYPE>::Node*
List<TYPE>::getHead() const
{
    return mHead;
}

template <typename TYPE>
typename List<TYPE>::iterator
List<TYPE>::getIterator()
{
    return typename List<TYPE>::iterator(mHead);
}

